@@ -2,6 +2,7 @@
 
 import logging
 import sys
+import warnings
 from argparse import Action, ArgumentError
 from datetime import datetime
 from difflib import unified_diff
@@ -10,7 +11,7 @@ from typing import Collection, Generator, List, Tuple
 
 from darker.black_diff import (
     BlackConfig,
-    apply_black_excludes,
+    filter_python_files,
     read_black_config,
     run_black,
 )
@@ -18,7 +19,9 @@ from darker.chooser import choose_lines
 from darker.command_line import parse_command_line
 from darker.config import OutputMode, dump_config
 from darker.diff import diff_and_get_opcodes, opcodes_to_chunks
+from darker.exceptions import DependencyError, MissingPackageError
 from darker.git import (
+    PRE_COMMIT_FROM_TO_REFS,
     WORKTREE,
     EditedLinenumsDiffer,
     RevisionRange,
@@ -29,15 +32,16 @@ from darker.git import (
 from darker.help import ISORT_INSTRUCTION
 from darker.import_sorting import apply_isort, isort
 from darker.linting import run_linters
-from darker.utils import GIT_DATEFORMAT, TextDocument, get_common_root
-from darker.verification import BinarySearch, NotEquivalentError, verify_ast_unchanged
+from darker.utils import GIT_DATEFORMAT, TextDocument, debug_dump, get_common_root
+from darker.verification import ASTVerifier, BinarySearch, NotEquivalentError
 
 logger = logging.getLogger(__name__)
 
 
 def format_edited_parts(
-    git_root: Path,
+    root: Path,
     changed_files: Collection[Path],  # pylint: disable=unsubscriptable-object
+    black_exclude: Collection[Path],  # pylint: disable=unsubscriptable-object
     revrange: RevisionRange,
     enable_isort: bool,
     black_config: BlackConfig,
@@ -45,13 +49,17 @@ def format_edited_parts(
 ) -> Generator[Tuple[Path, TextDocument, TextDocument], None, None]:
     """Black (and optional isort) formatting for chunks with edits since the last commit
 
-    Files excluded by Black's configuration are not reformatted using Black, but their
-    imports are still sorted. Also, linters will be run for all files in a separate step
-    after this function has completed.
+    Files inside given directories and excluded by Black's configuration are not
+    reformatted using Black, but their imports are still sorted. Also, linters will be
+    run for all files in a separate step after this function has completed.
 
-    :param git_root: The root of the Git repository the files are in
-    :param changed_files: Files which have been modified in the repository between the
-                          given Git revisions
+    Files listed explicitly on the command line are always reformatted.
+
+    :param root: The common root of all files to reformat
+    :param changed_files: Python files and explicitly requested files which have been
+                          modified in the repository between the given Git revisions
+    :param black_exclude: Python files to not reformat using Black, according to Black
+                          configuration
     :param revrange: The Git revisions to compare
     :param enable_isort: ``True`` to also run ``isort`` first on each changed file
     :param black_config: Configuration to use for running Black
@@ -60,12 +68,9 @@ def format_edited_parts(
              be reformatted, and skips unchanged files.
 
     """
-    files_to_blacken = apply_black_excludes(changed_files, git_root, black_config)
     for path_in_repo in sorted(changed_files):
-        src = git_root / path_in_repo
-        rev2_content = git_get_content_at_revision(
-            path_in_repo, revrange.rev2, git_root
-        )
+        src = root / path_in_repo
+        rev2_content = git_get_content_at_revision(path_in_repo, revrange.rev2, root)
 
         # 1. run isort
         if enable_isort:
@@ -77,12 +82,12 @@ def format_edited_parts(
             )
         else:
             rev2_isorted = rev2_content
-        if src in files_to_blacken:
+        if path_in_repo not in black_exclude:
             # 9. A re-formatted Python file which produces an identical AST was
             #    created successfully - write an updated file or print the diff if
             #    there were any changes to the original
             content_after_reformatting = _reformat_single_file(
-                git_root,
+                root,
                 path_in_repo,
                 revrange,
                 rev2_content,
@@ -98,8 +103,8 @@ def format_edited_parts(
 
 
 def _reformat_single_file(  # pylint: disable=too-many-arguments,too-many-locals
-    git_root: Path,
-    path_in_repo: Path,
+    root: Path,
+    relative_path: Path,
     revrange: RevisionRange,
     rev2_content: TextDocument,
     rev2_isorted: TextDocument,
@@ -108,8 +113,8 @@ def _reformat_single_file(  # pylint: disable=too-many-arguments,too-many-locals
 ) -> TextDocument:
     """In a Python file, reformat chunks with edits since the last commit using Black
 
-    :param git_root: The root of the Git repository the files are in
-    :param path_in_repo: Relative path to a Python source code file
+    :param root: The common root of all files to reformat
+    :param relative_path: Relative path to a Python source code file
     :param revrange: The Git revisions to compare
     :param rev2_content: Contents of the file at ``revrange.rev2``
     :param rev2_isorted: Contents of the file after optional import sorting
@@ -119,12 +124,31 @@ def _reformat_single_file(  # pylint: disable=too-many-arguments,too-many-locals
     :raise: NotEquivalentError
 
     """
-    src = git_root / path_in_repo
-    edited_linenums_differ = EditedLinenumsDiffer(git_root, revrange)
+    src = root / relative_path
+    rev1_relative_path = _get_rev1_path(relative_path)
+    edited_linenums_differ = EditedLinenumsDiffer(root, revrange)
+
+    # 4. run black
+    formatted = run_black(rev2_isorted, black_config)
+    logger.debug("Read %s lines from edited file %s", len(rev2_isorted.lines), src)
+    logger.debug("Black reformat resulted in %s lines", len(formatted.lines))
+
+    # 5. get the diff between the edited and reformatted file
+    opcodes = diff_and_get_opcodes(rev2_isorted, formatted)
+
+    # 6. convert the diff into chunks
+    black_chunks = list(opcodes_to_chunks(opcodes, rev2_isorted, formatted))
+
+    # Exit early if nothing to do
+    if not black_chunks:
+        return rev2_isorted
 
     max_context_lines = len(rev2_isorted.lines)
     minimum_context_lines = BinarySearch(0, max_context_lines + 1)
     last_successful_reformat = None
+
+    verifier = ASTVerifier(baseline=rev2_isorted)
+
     while not minimum_context_lines.found:
         context_lines = minimum_context_lines.get_next()
         if context_lines > 0:
@@ -136,24 +160,13 @@ def _reformat_single_file(  # pylint: disable=too-many-arguments,too-many-locals
         # 2. diff the given revision and worktree for the file
         # 3. extract line numbers in the edited to-file for changed lines
         edited_linenums = edited_linenums_differ.revision_vs_lines(
-            path_in_repo, rev2_isorted, context_lines
+            rev1_relative_path, rev2_isorted, context_lines
         )
         if enable_isort and not edited_linenums and rev2_isorted == rev2_content:
             logger.debug("No changes in %s after isort", src)
             last_successful_reformat = rev2_isorted
             break
 
-        # 4. run black
-        formatted = run_black(rev2_isorted, black_config)
-        logger.debug("Read %s lines from edited file %s", len(rev2_isorted.lines), src)
-        logger.debug("Black reformat resulted in %s lines", len(formatted.lines))
-
-        # 5. get the diff between the edited and reformatted file
-        opcodes = diff_and_get_opcodes(rev2_isorted, formatted)
-
-        # 6. convert the diff into chunks
-        black_chunks = list(opcodes_to_chunks(opcodes, rev2_isorted, formatted))
-
         # 7. choose reformatted content
         chosen = TextDocument.from_lines(
             choose_lines(black_chunks, edited_linenums),
@@ -169,13 +182,8 @@ def _reformat_single_file(  # pylint: disable=too-many-arguments,too-many-locals
             len(rev2_isorted.lines),
             len(chosen.lines),
         )
-        try:
-            verify_ast_unchanged(rev2_isorted, chosen, black_chunks, edited_linenums)
-        except NotEquivalentError:
-            # Diff produced misaligned chunks which couldn't be reconstructed into
-            # a partially re-formatted Python file which produces an identical AST.
-            # Try again with a larger `-U<context_lines>` option for `git diff`,
-            # or give up if `context_lines` is already very large.
+        if not verifier.is_equivalent_to_baseline(chosen):
+            debug_dump(black_chunks, edited_linenums)
             logger.debug(
                 "AST verification of %s with %s lines of context failed",
                 src,
@@ -186,27 +194,49 @@ def _reformat_single_file(  # pylint: disable=too-many-arguments,too-many-locals
             minimum_context_lines.respond(True)
             last_successful_reformat = chosen
     if not last_successful_reformat:
-        raise NotEquivalentError(path_in_repo)
+        raise NotEquivalentError(relative_path)
     return last_successful_reformat
 
 
+def _get_rev1_path(path: Path) -> Path:
+    """Return the relative path to the file in the old revision
+
+    This is usually the same as the relative path on the command line. But in the
+    special case of VSCode temporary files (like ``file.py.12345.tmp``), we actually
+    want to diff against the corresponding ``.py`` file instead.
+
+    """
+    if path.suffixes[-3::2] != [".py", ".tmp"]:
+        # The file name is not like `*.py.<HASH>.tmp`. Return it as such.
+        return path
+    # This is a VSCode temporary file. Drop the hash and the `.tmp` suffix to get the
+    # original file name for retrieving the previous revision to diff against.
+    path_with_hash = path.with_suffix("")
+    return path_with_hash.with_suffix("")
+
+
 def modify_file(path: Path, new_content: TextDocument) -> None:
     """Write new content to a file and inform the user by logging"""
     logger.info("Writing %s bytes into %s", len(new_content.string), path)
     path.write_bytes(new_content.encoded_string)
 
 
-def print_diff(path: Path, old: TextDocument, new: TextDocument) -> None:
+def print_diff(
+    path: Path, old: TextDocument, new: TextDocument, root: Path = None
+) -> None:
     """Print ``black --diff`` style output for the changes
 
-    :param path: The relative path of the file to print the diff output for
+    :param path: The path of the file to print the diff output for, relative to CWD
     :param old: Old contents of the file
     :param new: New contents of the file
+    :param root: The root for the relative path (current working directory if omitted)
 
     Modification times should be in the format "YYYY-MM-DD HH:MM:SS:mmmmmm +0000"
 
     """
-    relative_path = path.resolve().relative_to(Path.cwd()).as_posix()
+    if root is None:
+        root = Path.cwd()
+    relative_path = path.resolve().relative_to(root).as_posix()
     diff = "\n".join(
         line.rstrip("\n")
         for line in unified_diff(
@@ -296,8 +326,7 @@ def main(argv: List[str] = None) -> int:
         print("\n")
 
     if args.isort and not isort:
-        logger.error(f"{ISORT_INSTRUCTION} to use the `--isort` option.")
-        exit(1)
+        raise MissingPackageError(f"{ISORT_INSTRUCTION} to use the `--isort` option.")
 
     black_config = read_black_config(tuple(args.src), args.config)
     if args.config:
@@ -310,18 +339,24 @@ def main(argv: List[str] = None) -> int:
         black_config["skip_magic_trailing_comma"] = args.skip_magic_trailing_comma
 
     paths = {Path(p) for p in args.src}
-    git_root = get_common_root(paths)
-    failures_on_modified_lines = False
+    root = get_common_root(paths)
 
-    revrange = RevisionRange.parse(args.revision)
+    revrange = RevisionRange.parse_with_common_ancestor(args.revision, root)
     output_mode = OutputMode.from_args(args)
     write_modified_files = not args.check and output_mode == OutputMode.NOTHING
-    if revrange.rev2 != WORKTREE and write_modified_files:
-        raise ArgumentError(
-            Action(["-r", "--revision"], "revision"),
-            f"Can't write reformatted files for revision '{revrange.rev2}'."
-            " Either --diff or --check must be used.",
-        )
+    if write_modified_files:
+        if args.revision == PRE_COMMIT_FROM_TO_REFS and revrange.rev2 == "HEAD":
+            warnings.warn(
+                "Darker was called by pre-commit, comparing HEAD to an older commit."
+                " As an experimental feature, allowing overwriting of files."
+                " See https://github.com/akaihola/darker/issues/180 for details."
+            )
+        elif revrange.rev2 != WORKTREE:
+            raise ArgumentError(
+                Action(["-r", "--revision"], "revision"),
+                f"Can't write reformatted files for revision '{revrange.rev2}'."
+                " Either --diff or --check must be used.",
+            )
 
     missing = get_missing_at_revision(paths, revrange.rev2)
     if missing:
@@ -332,38 +367,61 @@ def main(argv: List[str] = None) -> int:
             f"Error: Path(s) {missing_reprs} do not exist in {rev2_repr}",
         )
 
+    # These are absolute paths:
+    files_to_process = filter_python_files(paths, root, {})
+    files_to_blacken = filter_python_files(paths, root, black_config)
     if output_mode == OutputMode.CONTENT:
         # With `-d` / `--stdout`, process the file whether modified or not. Paths have
         # previously been validated to contain exactly one existing file.
-        changed_files = paths
+        changed_files_to_process = {
+            p.resolve().relative_to(root) for p in files_to_process
+        }
+        black_exclude = set()
     else:
         # In other modes, only process files which have been modified.
-        changed_files = git_get_modified_files(paths, revrange, git_root)
+        # These are relative to `root`:
+        changed_files_to_process = git_get_modified_files(
+            files_to_process, revrange, root
+        )
+        black_exclude = {
+            f for f in changed_files_to_process if root / f not in files_to_blacken
+        }
+
+    formatting_failures_on_modified_lines = False
     for path, old, new in format_edited_parts(
-        git_root,
-        changed_files,
+        root,
+        changed_files_to_process,
+        black_exclude,
         revrange,
         args.isort,
         black_config,
         report_unmodified=output_mode == OutputMode.CONTENT,
     ):
-        failures_on_modified_lines = True
+        formatting_failures_on_modified_lines = True
         if output_mode == OutputMode.DIFF:
-            print_diff(path, old, new)
+            print_diff(path, old, new, root)
         elif output_mode == OutputMode.CONTENT:
             print_source(new)
         if write_modified_files:
             modify_file(path, new)
-    if run_linters(args.lint, git_root, changed_files, revrange):
-        failures_on_modified_lines = True
-    return 1 if args.check and failures_on_modified_lines else 0
+    linter_failures_on_modified_lines = run_linters(
+        args.lint, root, changed_files_to_process, revrange
+    )
+    return (
+        1
+        if linter_failures_on_modified_lines
+        or (args.check and formatting_failures_on_modified_lines)
+        else 0
+    )
 
 
 def main_with_error_handling() -> int:
     """Entry point for console script"""
     try:
         return main()
-    except ArgumentError as exc_info:
+    except (ArgumentError, DependencyError) as exc_info:
+        if logger.root.level < logging.WARNING:
+            raise
         sys.exit(str(exc_info))
 
 
