@@ -338,7 +338,7 @@ def test_help_description_without_isort_package(capsys):
     with isort_present(False):
 
         assert (
-            "Please run `pip install 'darker[isort]'` to enable sorting of import "
+            "Please run `pip install darker[isort]` to enable sorting of import "
             "definitions" in get_darker_help_output(capsys)
         )
 
@@ -348,7 +348,7 @@ def test_help_isort_option_without_isort_package(capsys):
     with isort_present(False):
 
         assert (
-            "Please run `pip install 'darker[isort]'` to enable usage of this option."
+            "Please run `pip install darker[isort]` to enable usage of this option."
             in get_darker_help_output(capsys)
         )
 
@@ -509,18 +509,33 @@ def test_black_options_skip_magic_trailing_comma(git_repo, config, options, expe
         options=["a.py"],
         # Expected arguments to the `format_edited_parts()` call.
         # `Path("git_root")` will be replaced with the temporary Git repository root:
-        expect=(Path("git_root"), {Path("a.py")}, RevisionRange("HEAD"), False, {}),
+        expect=(
+            Path("git_root"),
+            {Path("a.py")},
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
+            False,
+            {},
+        ),
     ),
     dict(
         options=["--isort", "a.py"],
-        expect=(Path("git_root"), {Path("a.py")}, RevisionRange("HEAD"), True, {}),
+        expect=(
+            Path("git_root"),
+            {Path("a.py")},
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
+            True,
+            {},
+        ),
     ),
     dict(
         options=["--config", "my.cfg", "a.py"],
         expect=(
             Path("git_root"),
             {Path("a.py")},
-            RevisionRange("HEAD"),
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
             False,
             {"config": "my.cfg"},
         ),
@@ -530,7 +545,8 @@ def test_black_options_skip_magic_trailing_comma(git_repo, config, options, expe
         expect=(
             Path("git_root"),
             {Path("a.py")},
-            RevisionRange("HEAD"),
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
             False,
             {"line_length": 90},
         ),
@@ -540,14 +556,22 @@ def test_black_options_skip_magic_trailing_comma(git_repo, config, options, expe
         expect=(
             Path("git_root"),
             {Path("a.py")},
-            RevisionRange("HEAD"),
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
             False,
             {"skip_string_normalization": True},
         ),
     ),
     dict(
         options=["--diff", "a.py"],
-        expect=(Path("git_root"), {Path("a.py")}, RevisionRange("HEAD"), False, {}),
+        expect=(
+            Path("git_root"),
+            {Path("a.py")},
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
+            False,
+            {},
+        ),
     ),
 )
 def test_options(git_repo, options, expect):
@@ -570,14 +594,16 @@ def test_options(git_repo, options, expect):
 
 
 @pytest.mark.kwparametrize(
-    dict(check=False, changes=False),
-    dict(check=False, changes=True),
-    dict(check=True, changes=False),
-    dict(check=True, changes=True, expect_retval=1),
+    dict(check=False, changes=False, lintfail=False),
+    dict(check=False, changes=False, lintfail=True, expect_retval=1),
+    dict(check=False, changes=True, lintfail=False),
+    dict(check=False, changes=True, lintfail=True, expect_retval=1),
+    dict(check=True, changes=False, lintfail=False),
+    dict(check=True, changes=True, lintfail=True, expect_retval=1),
     expect_retval=0,
 )
-def test_main_retval(git_repo, check, changes, expect_retval):
-    """main() return value is correct based on --check and the need to reformat files"""
+def test_main_retval(git_repo, check, changes, lintfail, expect_retval):
+    """main() return value is correct based on --check, reformatting and linters"""
     git_repo.add({"a.py": ""}, commit="Initial commit")
     format_edited_parts = Mock()
     format_edited_parts.return_value = (
@@ -591,9 +617,13 @@ def test_main_retval(git_repo, check, changes, expect_retval):
         if changes
         else []
     )
+    run_linters = Mock(return_value=lintfail)
     check_arg_maybe = ["--check"] if check else []
     with patch.multiple(
-        "darker.__main__", format_edited_parts=format_edited_parts, modify_file=DEFAULT
+        "darker.__main__",
+        format_edited_parts=format_edited_parts,
+        modify_file=DEFAULT,
+        run_linters=run_linters,
     ):
 
         retval = main(check_arg_maybe + ["a.py"])
