@@ -8,7 +8,7 @@ from datetime import datetime, timedelta
 from pathlib import Path
 from subprocess import DEVNULL, PIPE, CalledProcessError, check_call
 from typing import List, Union
-from unittest.mock import patch
+from unittest.mock import call, patch
 
 import pytest
 
@@ -115,50 +115,108 @@ def test_git_get_content_at_revision(git_repo, revision, expect_lines, expect_mt
 )
 def test_revisionrange_parse(revision_range, expect):
     """Test for :meth:`RevisionRange.parse`"""
-    revrange = git.RevisionRange.parse(revision_range)
-    assert (revrange.rev1, revrange.rev2, revrange.use_common_ancestor) == expect
+    result = git.RevisionRange._parse(revision_range)
+
+    assert result == expect
 
 
 @pytest.mark.kwparametrize(
+    dict(
+        revision=":WORKTREE:",
+        expect_textdocument_calls=[call.from_file(Path("/path/my.txt"))],
+    ),
     dict(
         revision="HEAD",
-        expect=[
+        expect_git_calls=[
             "git show HEAD:./my.txt",
             "git log -1 --format=%ct HEAD -- my.txt",
         ],
+        expect_textdocument_calls=[
+            call.from_lines([b"1627107028"], mtime="2021-07-24 06:10:28.000000 +0000")
+        ],
     ),
     dict(
         revision="HEAD^",
-        expect=[
+        expect_git_calls=[
             "git show HEAD^:./my.txt",
             "git log -1 --format=%ct HEAD^ -- my.txt",
         ],
+        expect_textdocument_calls=[
+            call.from_lines([b"1627107028"], mtime="2021-07-24 06:10:28.000000 +0000")
+        ],
     ),
     dict(
         revision="master",
-        expect=[
+        expect_git_calls=[
             "git show master:./my.txt",
             "git log -1 --format=%ct master -- my.txt",
         ],
+        expect_textdocument_calls=[
+            call.from_lines([b"1627107028"], mtime="2021-07-24 06:10:28.000000 +0000")
+        ],
     ),
+    expect_git_calls=[],
 )
-def test_git_get_content_at_revision_git_calls(revision, expect):
-    """get_git_content_at_revision() calls Git correctly"""
-    with patch("darker.git.check_output") as check_output:
-        # a dummy Unix timestamp:
-        check_output.return_value = b"1000000"
+def test_git_get_content_at_revision_obtain_file_content(
+    revision, expect_git_calls, expect_textdocument_calls
+):
+    """``git_get_content_at_revision`` calls Git or reads files based on revision"""
+    with patch("darker.git.check_output") as check_output, patch(
+        "darker.git.TextDocument"
+    ) as text_document_class:
+        # this dummy value acts both as a dummy Unix timestamp for the file as well as
+        # the contents of the file:
+        check_output.return_value = b"1627107028"
 
-        git.git_get_content_at_revision(Path("my.txt"), revision, Path("cwd"))
+        git.git_get_content_at_revision(Path("my.txt"), revision, Path("/path"))
 
-        assert check_output.call_count == len(expect)
-        for expect_call in expect:
-            check_output.assert_any_call(
-                expect_call.split(),
-                cwd="cwd",
+        expected_calls = [
+            call(
+                expected_call.split(),
+                cwd=str(Path("/path")),
                 encoding="utf-8",
                 stderr=PIPE,
                 env={"LC_ALL": "C"},
             )
+            for expected_call in expect_git_calls
+        ]
+        assert check_output.call_args_list == expected_calls
+        assert text_document_class.method_calls == expect_textdocument_calls
+
+
+@pytest.mark.kwparametrize(
+    dict(revrange="HEAD", expect="HEAD..:WORKTREE:"),
+    dict(revrange="{initial}", expect="{initial}..:WORKTREE:"),
+    dict(revrange="{initial}..", expect="{initial}..:WORKTREE:"),
+    dict(revrange="{initial}..HEAD", expect="{initial}..HEAD"),
+    dict(revrange="{initial}..feature", expect="{initial}..feature"),
+    dict(revrange="{initial}...", expect="{initial}..:WORKTREE:"),
+    dict(revrange="{initial}...HEAD", expect="{initial}..HEAD"),
+    dict(revrange="{initial}...feature", expect="{initial}..feature"),
+    dict(revrange="master", expect="{initial}..:WORKTREE:"),
+    dict(revrange="master..", expect="master..:WORKTREE:"),
+    dict(revrange="master..HEAD", expect="master..HEAD"),
+    dict(revrange="master..feature", expect="master..feature"),
+    dict(revrange="master...", expect="{initial}..:WORKTREE:"),
+    dict(revrange="master...HEAD", expect="{initial}..HEAD"),
+    dict(revrange="master...feature", expect="{initial}..feature"),
+)
+def test_revisionrange_parse_with_common_ancestor(git_repo, revrange, expect):
+    """``_git_get_old_revision()`` gets common ancestor using Git when necessary"""
+    git_repo.add({"a": "i"}, commit="Initial commit")
+    initial = git_repo.get_hash()
+    git_repo.add({"a": "m"}, commit="in master")
+    master = git_repo.get_hash()
+    git_repo.create_branch("feature", initial)
+    git_repo.add({"a": "f"}, commit="in feature")
+
+    result = git.RevisionRange.parse_with_common_ancestor(
+        revrange.format(initial=initial), git_repo.root
+    )
+
+    rev1, rev2 = expect.format(initial=initial, master=master).split("..")
+    assert result.rev1 == rev1
+    assert result.rev2 == rev2
 
 
 @pytest.mark.kwparametrize(
@@ -391,6 +449,48 @@ def test_get_missing_at_revision_worktree(git_repo):
     assert result == {Path("dir/a.py"), Path("dir/b.py")}
 
 
+def test_git_diff_name_only(git_repo):
+    """``_git_diff_name_only()`` only returns paths of modified files"""
+    git_repo.add({"a.py": "a", "b.py": "b", "c.py": "c"}, commit="Initial commit")
+    first = git_repo.get_hash()
+    git_repo.add({"a.py": "A", "b.dy": "B"}, commit="only a.py modified")
+    second = git_repo.get_hash()
+
+    result = git._git_diff_name_only(
+        first, second, {Path("a.py"), Path("c.py"), Path("Z.py")}, git_repo.root
+    )
+
+    assert result == {Path("a.py")}
+
+
+def test_git_ls_files_others(git_repo):
+    """``_git_ls_files_others()`` only returns paths of untracked non-ignored files"""
+    git_repo.add(
+        {
+            "tracked.py": "tracked",
+            "tracked.ignored": "tracked",
+            ".gitignore": "*.ignored",
+        },
+        commit="Initial commit",
+    )
+    (git_repo.root / "untracked.py").write_text("untracked")
+    (git_repo.root / "untracked.ignored").write_text("untracked")
+
+    result = git._git_ls_files_others(
+        {
+            Path("tracked.py"),
+            Path("tracked.ignored"),
+            Path("untracked.py"),
+            Path("untracked.ignored"),
+            Path("missing.py"),
+            Path("missing.ignored"),
+        },
+        git_repo.root,
+    )
+
+    assert result == {Path("untracked.py")}
+
+
 @pytest.mark.kwparametrize(
     dict(paths=["a.py"], expect=[]),
     dict(expect=[]),
@@ -428,10 +528,9 @@ def test_git_get_modified_files(git_repo, modify_paths, paths, expect):
         else:
             absolute_path.parent.mkdir(parents=True, exist_ok=True)
             absolute_path.write_bytes(content.encode("ascii"))
+    revrange = git.RevisionRange("HEAD", ":WORKTREE:")
 
-    result = git.git_get_modified_files(
-        {root / p for p in paths}, git.RevisionRange("HEAD"), cwd=root
-    )
+    result = git.git_get_modified_files({root / p for p in paths}, revrange, cwd=root)
 
     assert result == {Path(p) for p in expect}
 
@@ -584,7 +683,7 @@ def test_git_get_modified_files_revision_range(
     """Test for :func:`darker.git.git_get_modified_files` with a revision range"""
     result = git.git_get_modified_files(
         [Path(branched_repo.root)],
-        git.RevisionRange.parse(revrange),
+        git.RevisionRange.parse_with_common_ancestor(revrange, branched_repo.root),
         Path(branched_repo.root),
     )
 
@@ -620,14 +719,12 @@ def test_git_get_modified_files_revision_range(
 def test_revisionrange_parse_pre_commit(
     environ, expect_rev1, expect_rev2, expect_use_common_ancestor
 ):
-    """RevisionRange.parse(':PRE-COMMIT:') gets the range from environment variables"""
+    """RevisionRange._parse(':PRE-COMMIT:') gets the range from environment variables"""
     with patch.dict(os.environ, environ):
 
-        result = git.RevisionRange.parse(":PRE-COMMIT:")
+        result = git.RevisionRange._parse(":PRE-COMMIT:")
 
-        assert result.rev1 == expect_rev1
-        assert result.rev2 == expect_rev2
-        assert result.use_common_ancestor == expect_use_common_ancestor
+        assert result == (expect_rev1, expect_rev2, expect_use_common_ancestor)
 
 
 edited_linenums_differ_cases = pytest.mark.kwparametrize(
@@ -643,7 +740,8 @@ def test_edited_linenums_differ_compare_revisions(git_repo, context_lines, expec
     """Tests for EditedLinenumsDiffer.revision_vs_worktree()"""
     paths = git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
     paths["a.py"].write_bytes(b"1\n2\nthree\n4\n5\n6\nseven\n8\n")
-    differ = git.EditedLinenumsDiffer(Path(git_repo.root), git.RevisionRange("HEAD"))
+    revrange = git.RevisionRange("HEAD", ":WORKTREE:")
+    differ = git.EditedLinenumsDiffer(git_repo.root, revrange)
 
     linenums = differ.compare_revisions(Path("a.py"), context_lines)
 
@@ -655,7 +753,8 @@ def test_edited_linenums_differ_revision_vs_lines(git_repo, context_lines, expec
     """Tests for EditedLinenumsDiffer.revision_vs_lines()"""
     git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
     content = TextDocument.from_lines(["1", "2", "three", "4", "5", "6", "seven", "8"])
-    differ = git.EditedLinenumsDiffer(git_repo.root, git.RevisionRange("HEAD"))
+    revrange = git.RevisionRange("HEAD", ":WORKTREE:")
+    differ = git.EditedLinenumsDiffer(git_repo.root, revrange)
 
     linenums = differ.revision_vs_lines(Path("a.py"), content, context_lines)
 
