@@ -1,6 +1,6 @@
 """Unit tests for :mod:`darker.__main__`"""
 
-# pylint: disable=unused-argument
+# pylint: disable=unused-argument,too-many-arguments,use-dict-literal,protected-access
 
 import logging
 import re
@@ -8,13 +8,14 @@ from argparse import ArgumentError
 from pathlib import Path
 from textwrap import dedent
 from types import SimpleNamespace
-from unittest.mock import Mock, patch
+from unittest.mock import patch
 
 import pytest
 from black import find_project_root
 
 import darker.__main__
 import darker.import_sorting
+from darker.exceptions import MissingPackageError
 from darker.git import WORKTREE, RevisionRange
 from darker.tests.helpers import isort_present
 from darker.utils import TextDocument, joinlines
@@ -25,13 +26,13 @@ def test_isort_option_without_isort(git_repo, caplog):
     """Without isort, provide isort install instructions and error"""
     with isort_present(False), patch.object(
         darker.__main__, "isort", None
-    ), pytest.raises(SystemExit):
+    ), pytest.raises(MissingPackageError) as exc_info:
 
         darker.__main__.main(["--isort", "."])
 
     assert (
-        "Please run `pip install 'darker[isort]'` to use the `--isort` option."
-        in caplog.text
+        str(exc_info.value)
+        == "Please run `pip install darker[isort]` to use the `--isort` option."
     )
 
 
@@ -43,12 +44,13 @@ def run_isort(git_repo, monkeypatch, caplog, request):
     paths = git_repo.add({"test1.py": "original"}, commit="Initial commit")
     paths["test1.py"].write_bytes(b"changed")
     args = getattr(request, "param", ())
-    with patch.multiple(
-        darker.__main__,
-        run_black=Mock(return_value=TextDocument()),
-        verify_ast_unchanged=Mock(),
-    ), patch("darker.import_sorting.isort_code") as isort_code:
-        isort_code.return_value = "dummy isort output"
+    isorted_code = "import os; import sys;"
+    blacken_code = "import os\nimport sys\n"
+    patch_run_black_ctx = patch.object(
+        darker.__main__, "run_black", return_value=TextDocument(blacken_code)
+    )
+    with patch_run_black_ctx, patch("darker.import_sorting.isort_code") as isort_code:
+        isort_code.return_value = isorted_code
         darker.__main__.main(["--isort", "./test1.py", *args])
         return SimpleNamespace(
             isort_code=darker.import_sorting.isort_code, caplog=caplog
@@ -112,49 +114,45 @@ A_PY_DIFF_BLACK_ISORT = [
 
 
 @pytest.mark.kwparametrize(
-    dict(enable_isort=False, black_config={}, expect=[A_PY_BLACK]),
-    dict(enable_isort=True, black_config={}, expect=[A_PY_BLACK_ISORT]),
+    dict(
+        enable_isort=False,
+        black_config={},
+        black_exclude=set(),
+        expect=[A_PY_BLACK],
+    ),
+    dict(
+        enable_isort=True,
+        black_config={},
+        black_exclude=set(),
+        expect=[A_PY_BLACK_ISORT],
+    ),
     dict(
         enable_isort=False,
         black_config={"skip_string_normalization": True},
+        black_exclude=set(),
         expect=[A_PY_BLACK_UNNORMALIZE],
     ),
     dict(
         enable_isort=False,
-        black_config={"exclude": re.compile(r"/a\.py$")},
+        black_exclude={Path("a.py")},
         expect=[],
     ),
     dict(
         enable_isort=True,
-        black_config={"exclude": re.compile(r"/a\.py$")},
-        expect=[A_PY_ISORT],
-    ),
-    dict(
-        enable_isort=False,
-        black_config={"extend_exclude": re.compile(r"/a\.py$")},
-        expect=[],
-    ),
-    dict(
-        enable_isort=True,
-        black_config={"extend_exclude": re.compile(r"/a\.py$")},
-        expect=[A_PY_ISORT],
-    ),
-    dict(
-        enable_isort=False,
-        black_config={"force_exclude": re.compile(r"/a\.py$")},
-        expect=[],
-    ),
-    dict(
-        enable_isort=True,
-        black_config={"force_exclude": re.compile(r"/a\.py$")},
+        black_exclude={Path("a.py")},
         expect=[A_PY_ISORT],
     ),
+    black_config={},
 )
 @pytest.mark.parametrize("newline", ["\n", "\r\n"], ids=["unix", "windows"])
-def test_format_edited_parts(git_repo, enable_isort, black_config, newline, expect):
+def test_format_edited_parts(
+    git_repo, enable_isort, black_config, black_exclude, newline, expect
+):
     """Correct reformatting and import sorting changes are produced
 
-    Also, Black reformatting is skipped for files excluded in Black configuration
+    Black reformatting is done even if a file is excluded in Black configuration.
+    File exclusion is done in Darker before calling
+    :func:`~darker.__main__.format_edited_parts`.
 
     """
     paths = git_repo.add({"a.py": newline, "b.py": newline}, commit="Initial commit")
@@ -163,8 +161,9 @@ def test_format_edited_parts(git_repo, enable_isort, black_config, newline, expe
 
     result = darker.__main__.format_edited_parts(
         Path(git_repo.root),
-        [Path("a.py")],
-        RevisionRange("HEAD"),
+        {Path("a.py")},
+        black_exclude,
+        RevisionRange("HEAD", ":WORKTREE:"),
         enable_isort,
         black_config,
         report_unmodified=False,
@@ -197,7 +196,8 @@ def test_format_edited_parts_all_unchanged(git_repo, monkeypatch):
         darker.__main__.format_edited_parts(
             Path(git_repo.root),
             {Path("a.py"), Path("b.py")},
-            RevisionRange("HEAD"),
+            set(),
+            RevisionRange("HEAD", ":WORKTREE:"),
             True,
             {},
             report_unmodified=False,
@@ -212,16 +212,16 @@ def test_format_edited_parts_ast_changed(git_repo, caplog):
     caplog.set_level(logging.DEBUG, logger="darker.__main__")
     paths = git_repo.add({"a.py": "1\n2\n3\n4\n5\n6\n7\n8\n"}, commit="Initial commit")
     paths["a.py"].write_bytes(b"8\n7\n6\n5\n4\n3\n2\n1\n")
-    with patch.object(
-        darker.__main__, "verify_ast_unchanged"
-    ) as verify_ast_unchanged, pytest.raises(NotEquivalentError):
-        verify_ast_unchanged.side_effect = NotEquivalentError
-
+    mock_ctx = patch.object(
+        darker.verification.ASTVerifier, "is_equivalent_to_baseline", return_value=False
+    )
+    with mock_ctx, pytest.raises(NotEquivalentError):
         _ = list(
             darker.__main__.format_edited_parts(
                 git_repo.root,
-                [Path("a.py")],
-                RevisionRange("HEAD"),
+                {Path("a.py")},
+                set(),
+                RevisionRange("HEAD", ":WORKTREE:"),
                 enable_isort=False,
                 black_config={},
                 report_unmodified=False,
@@ -265,7 +265,8 @@ def test_format_edited_parts_isort_on_already_formatted(git_repo):
     result = darker.__main__.format_edited_parts(
         git_repo.root,
         {Path("a.py")},
-        RevisionRange("HEAD"),
+        set(),
+        RevisionRange("HEAD", ":WORKTREE:"),
         enable_isort=True,
         black_config={},
         report_unmodified=False,
@@ -320,6 +321,7 @@ def test_format_edited_parts_historical(git_repo, rev1, rev2, expect):
     result = darker.__main__.format_edited_parts(
         git_repo.root,
         {Path("a.py")},
+        set(),
         RevisionRange(rev1, rev2),
         enable_isort=True,
         black_config={},
@@ -329,6 +331,85 @@ def test_format_edited_parts_historical(git_repo, rev1, rev2, expect):
     assert list(result) == [(paths["a.py"], a_py[x[0]], a_py[x[1]]) for x in expect]
 
 
+@pytest.mark.kwparametrize(
+    dict(),
+    dict(relative_path="file.py.12345.tmp"),
+    dict(
+        rev2_content="import  modified\n\nprint( original )\n",
+        rev2_isorted="import  modified\n\nprint( original )\n",
+        expect="import modified\n\nprint( original )\n",
+    ),
+    dict(
+        rev2_content="import  original\n\nprint(modified )\n",
+        rev2_isorted="import  original\n\nprint(modified )\n",
+        expect="import  original\n\nprint(modified)\n",
+    ),
+    dict(
+        relative_path="file.py.12345.tmp",
+        rev2_content="import  modified\n\nprint( original )\n",
+        rev2_isorted="import  modified\n\nprint( original )\n",
+        expect="import modified\n\nprint( original )\n",
+    ),
+    dict(
+        relative_path="file.py.12345.tmp",
+        rev2_content="import  original\n\nprint(modified )\n",
+        rev2_isorted="import  original\n\nprint(modified )\n",
+        expect="import  original\n\nprint(modified)\n",
+    ),
+    relative_path="file.py",
+    rev1="HEAD",
+    rev2=":WORKTREE",
+    rev2_content="import  original\nprint( original )\n",
+    rev2_isorted="import  original\nprint( original )\n",
+    enable_isort=False,
+    black_config={},
+    expect="import  original\nprint( original )\n",
+)
+def test_reformat_single_file(
+    git_repo,
+    relative_path,
+    rev1,
+    rev2,
+    rev2_content,
+    rev2_isorted,
+    enable_isort,
+    black_config,
+    expect,
+):
+    """Test for ``_reformat_single_file``"""
+    git_repo.add(
+        {"file.py": "import  original\nprint( original )\n"}, commit="Initial commit"
+    )
+    result = darker.__main__._reformat_single_file(
+        git_repo.root,
+        Path(relative_path),
+        RevisionRange(rev1, rev2),
+        TextDocument(rev2_content),
+        TextDocument(rev2_isorted),
+        enable_isort,
+        black_config,
+    )
+
+    assert result.string == expect
+
+
+@pytest.mark.kwparametrize(
+    dict(path="file.py", expect="file.py"),
+    dict(path="subdir/file.py", expect="subdir/file.py"),
+    dict(path="file.py.12345.tmp", expect="file.py"),
+    dict(path="subdir/file.py.12345.tmp", expect="subdir/file.py"),
+    dict(path="file.py.tmp", expect="file.py.tmp"),
+    dict(path="subdir/file.py.tmp", expect="subdir/file.py.tmp"),
+    dict(path="file.12345.tmp", expect="file.12345.tmp"),
+    dict(path="subdir/file.12345.tmp", expect="subdir/file.12345.tmp"),
+)
+def test_get_rev1_path(path, expect):
+    """``_get_rev1_path`` drops two suffixes from ``.py.<HASH>.tmp``"""
+    result = darker.__main__._get_rev1_path(Path(path))
+
+    assert result == Path(expect)
+
+
 @pytest.mark.kwparametrize(
     dict(arguments=["--diff"], expect_stdout=A_PY_DIFF_BLACK),
     dict(arguments=["--isort"], expect_a_py=A_PY_BLACK_ISORT),
@@ -351,15 +432,17 @@ def test_format_edited_parts_historical(git_repo, rev1, rev2, expect):
         expect_retval=1,
     ),
     dict(
-        arguments=["--check", "--lint", "echo a.py:1: message"],
+        arguments=["--check", "--lint", "echo subdir/a.py:1: message"],
         # Windows compatible path assertion using `pathlib.Path()`
-        expect_stdout=[f"a.py:1: message {Path('/a.py')}"],
+        expect_stdout=[f"subdir/a.py:1: message {Path('/subdir/a.py')}"],
         expect_retval=1,
     ),
     dict(
-        arguments=["--diff", "--lint", "echo a.py:1: message"],
+        arguments=["--diff", "--lint", "echo subdir/a.py:1: message"],
         # Windows compatible path assertion using `pathlib.Path()`
-        expect_stdout=A_PY_DIFF_BLACK + [f"a.py:1: message {Path('/a.py')}"],
+        expect_stdout=A_PY_DIFF_BLACK
+        + [f"subdir/a.py:1: message {Path('/subdir/a.py')}"],
+        expect_retval=1,
     ),
     dict(
         arguments=[],
@@ -409,16 +492,23 @@ def test_format_edited_parts_historical(git_repo, rev1, rev2, expect):
            """,
         expect_stdout=[],
     ),
+    dict(
+        arguments=["--diff"],
+        expect_stdout=A_PY_DIFF_BLACK,
+        root_as_cwd=False,
+    ),
     # for all test cases, by default there's no output, `a.py` stays unmodified, and the
     # return value is a zero:
     pyproject_toml="",
     expect_stdout=[],
     expect_a_py=A_PY,
     expect_retval=0,
+    root_as_cwd=True,
 )
 @pytest.mark.parametrize("newline", ["\n", "\r\n"], ids=["unix", "windows"])
 def test_main(
     git_repo,
+    monkeypatch,
     capsys,
     find_project_root_cache_clear,
     arguments,
@@ -427,16 +517,29 @@ def test_main(
     expect_stdout,
     expect_a_py,
     expect_retval,
-):  # pylint: disable=too-many-arguments
+    root_as_cwd,
+    tmp_path_factory,
+):
     """Main function outputs diffs and modifies files correctly"""
+    if root_as_cwd:
+        cwd = git_repo.root
+        pwd = Path("")
+    else:
+        cwd = tmp_path_factory.mktemp("not_a_git_repo")
+        pwd = git_repo.root
+    monkeypatch.chdir(cwd)
     paths = git_repo.add(
-        {"pyproject.toml": dedent(pyproject_toml), "a.py": newline, "b.py": newline},
+        {
+            "pyproject.toml": dedent(pyproject_toml),
+            "subdir/a.py": newline,
+            "b.py": newline,
+        },
         commit="Initial commit",
     )
-    paths["a.py"].write_bytes(newline.join(A_PY).encode("ascii"))
+    paths["subdir/a.py"].write_bytes(newline.join(A_PY).encode("ascii"))
     paths["b.py"].write_bytes(f"print(42 ){newline}".encode("ascii"))
 
-    retval = darker.__main__.main(arguments + ["a.py"])
+    retval = darker.__main__.main(arguments + [str(pwd / "subdir")])
 
     stdout = capsys.readouterr().out.replace(str(git_repo.root), "")
     diff_output = stdout.splitlines(False)
@@ -452,7 +555,9 @@ def test_main(
         else:
             assert all("\t" not in line for line in diff_output)
     assert diff_output == expect_stdout
-    assert paths["a.py"].read_bytes().decode("ascii") == newline.join(expect_a_py)
+    assert paths["subdir/a.py"].read_bytes().decode("ascii") == newline.join(
+        expect_a_py
+    )
     assert paths["b.py"].read_bytes().decode("ascii") == f"print(42 ){newline}"
     assert retval == expect_retval
 
@@ -484,6 +589,49 @@ def test_main_historical(git_repo):
         darker.__main__.main(["--revision=foo..bar", "."])
 
 
+def test_main_pre_commit_head(git_repo, monkeypatch):
+    """Warn if run by pre-commit, rev2=HEAD and no ``--diff`` or ``--check`` provided"""
+    git_repo.add({"a.py": "original = 1"}, commit="Add a.py")
+    initial = git_repo.get_hash()
+    git_repo.add({"a.py": "modified  = 2"}, commit="Modify a.py")
+    monkeypatch.setenv("PRE_COMMIT_FROM_REF", initial)
+    monkeypatch.setenv("PRE_COMMIT_TO_REF", "HEAD")
+    with pytest.warns(
+        UserWarning,
+        match=re.escape(
+            "Darker was called by pre-commit, comparing HEAD to an older commit."
+            " As an experimental feature, allowing overwriting of files."
+            " See https://github.com/akaihola/darker/issues/180 for details."
+        ),
+    ):
+
+        result = darker.__main__.main(["--revision=:PRE-COMMIT:", "a.py"])
+
+    assert result == 0
+
+
+def test_main_historical_pre_commit(git_repo, monkeypatch):
+    """Stop if run by pre-commit, rev2 older than HEAD and no ``--diff``/``--check``"""
+    git_repo.add({"README.txt": ""}, commit="Initial commit")
+    initial = git_repo.get_hash()
+    git_repo.add({"a.py": "original"}, commit="Add a.py")
+    older_commit = git_repo.get_hash()
+    git_repo.add({"a.py": "modified"}, commit="Modify a.py")
+    monkeypatch.setenv("PRE_COMMIT_FROM_REF", initial)
+    monkeypatch.setenv("PRE_COMMIT_TO_REF", older_commit)
+    with pytest.raises(
+        ArgumentError,
+        match=(
+            re.escape(
+                f"Can't write reformatted files for revision '{older_commit}'."
+                " Either --diff or --check must be used."
+            )
+        ),
+    ):
+
+        darker.__main__.main(["--revision=:PRE-COMMIT:", "a.py"])
+
+
 def test_print_diff(tmp_path, monkeypatch, capsys):
     """print_diff() prints Black-style diff output with 5 lines of context"""
     monkeypatch.chdir(tmp_path)
@@ -573,3 +721,13 @@ def test_modify_file(tmp_path, new_content, expect):
 
     result = path.read_bytes()
     assert result == expect
+
+
+def test_stdout_path_resolution(git_repo, capsys):
+    """When using ``--stdout``, file paths are resolved correctly"""
+    git_repo.add({"src/menu.py": "print ( 'foo' )\n"})
+
+    result = darker.__main__.main(["--stdout", "./src/menu.py"])
+
+    assert result == 0
+    assert capsys.readouterr().out == 'print("foo")\n\n'
